@@ -194,6 +194,7 @@ class PandasDataSource:
                 self.renames[oldcol] = col
         self.df.columns = newcols
         self.pre_update = lambda df: df
+        self.post_update = lambda df: df
         self._period = None
         self.is_sparse = self.df[self.df.columns[self.col_data_offset]].isnull().sum() > len(self.df)/2
 
@@ -307,6 +308,7 @@ class PandasDataSource:
         for col in df.columns:
             if col not in input_df.columns:
                 input_df[col] = df[col]
+        input_df = self.post_update(input_df)
         input_df = input_df.reset_index()
         self.df = input_df[[input_df.columns[0]]+orig_cols] if orig_cols else input_df
         self.init_x1 = self.xlen - side_margin
@@ -350,9 +352,6 @@ class PandasDataSource:
         df = self.df.loc[x0:x1, :]
         if self.is_sparse:
             df = df.loc[df.iloc[:,self.col_data_offset].notna(), :]
-            if len(df) == 0:
-                # we need something that renders
-                df = self.df.loc[self.df.iloc[:,self.col_data_offset].notna(), :]
         origlen = len(df)
         return self._rows(df, colcnt, yscale=yscale, lod=lod), origlen
 
@@ -467,7 +466,7 @@ class FinCrossHair:
             for info in self.infos:
                 xtext,ytext = info(x,y,xtext,ytext)
         except Exception as e:
-            print(e)
+            print('Crosshair error:', type(e), e)
         space = '      '
         if close2right:
             xtext = xtext + space
@@ -715,6 +714,7 @@ class FinViewBox(pg.ViewBox):
             p0 = self.mapToView(ev.lastPos())
             p0 = _clamp_point(self.parent(), p0)
             s = nonzerosize(p0, p1)
+            p0 = QtCore.QPointF(p0.x()-s.x()/2, p0.y()-s.y()/2)
             self.draw_ellipse = FinEllipse(p0, s, pen=pg.mkPen(draw_line_color), movable=True)
             self.draw_ellipse.setZValue(80)
             self.rois.append(self.draw_ellipse)
@@ -930,12 +930,15 @@ class FinPlotItem(pg.GraphicsObject):
         self.painter = QtGui.QPainter()
         self.dirty = True
         self.lod = lod
+        self.cachedRect = None
 
     def repaint(self):
         self.dirty = True
         self.paint(self.painter)
 
     def paint(self, p, *args):
+        if self.datasrc.is_sparse:
+            self.dirty = True
         self.update_dirty_picture(self.viewRect())
         p.drawPicture(0, 0, self.picture)
 
@@ -949,10 +952,22 @@ class FinPlotItem(pg.GraphicsObject):
 
     def _generate_picture(self, boundingRect):
         w = boundingRect.width()
-        self.cachedRect = QtCore.QRectF(boundingRect.left()-w, 0, cache_candle_factor*w, 0)
+        self.cachedRect = QtCore.QRectF(boundingRect.left()-(cache_candle_factor-1)*0.5*w, 0, cache_candle_factor*w, 0)
+        self.painter.begin(self.picture)
+        self._generate_dummy_picture(self.viewRect())
         self.generate_picture(self.cachedRect)
+        self.painter.end()
         self.dirty = False
 
+    def _generate_dummy_picture(self, boundingRect):
+        if self.datasrc.is_sparse:
+            # just draw something to ensure PyQt will paint us again
+            self.painter.setPen(pg.mkPen(background))
+            self.painter.setBrush(pg.mkBrush(background))
+            l,r = boundingRect.left(), boundingRect.right()
+            self.painter.drawRect(QtCore.QRectF(l, boundingRect.top(), 1e-3, boundingRect.height()*1e-5))
+            self.painter.drawRect(QtCore.QRectF(r, boundingRect.bottom(), -1e-3, -boundingRect.height()*1e-5))
+
     def boundingRect(self):
         return QtCore.QRectF(self.picture.boundingRect())
 
@@ -985,7 +1000,6 @@ class CandlestickItem(FinPlotItem):
         w2 = w * 0.5
         left,right = boundingRect.left(), boundingRect.right()
         p = self.painter
-        p.begin(self.picture)
         df,origlen = self.datasrc.rows(5, left, right, yscale=self.ax.vb.yscale)
         drawing_many_shadows = self.draw_shadow and origlen > lod_candles*2//3
         for shadow,frame,body,df_rows in self.colorfunc(self, self.datasrc, df):
@@ -1003,7 +1017,6 @@ class CandlestickItem(FinPlotItem):
                 p.setBrush(pg.mkBrush(body))
                 for x,(t,open,close,high,low) in zip(idxs, rows):
                     p.drawRect(QtCore.QRectF(x-w2, open, w, close-open))
-        p.end()
 
     def rowcolors(self, prefix):
         return [self.colors[prefix+'_shadow'], self.colors[prefix+'_frame'], self.colors[prefix+'_body']]
@@ -1032,7 +1045,6 @@ class HeatmapItem(FinPlotItem):
         values /= np.nanmax(values) / (1+self.whiteout) # overshoot for coloring
         lim = self.filter_limit * (1+self.whiteout)
         p = self.painter
-        p.begin(self.picture)
         for t,row in enumerate(values):
             for ci,price in enumerate(prices):
                 v = row[ci]
@@ -1040,7 +1052,6 @@ class HeatmapItem(FinPlotItem):
                     v = 1 - self.colcurve(1 - (v-lim)/(1-lim))
                     color = self.colmap.map(v, mode='qcolor')
                     p.fillRect(QtCore.QRectF(t-rect_size2, price+h0, self.rect_size, h1), color)
-        p.end()
 
 
 
@@ -1081,7 +1092,6 @@ class HorizontalTimeVolumeItem(CandlestickItem):
         divvol[divvol==0] = 1
         volumes = (volumes * f / divvol).T
         p = self.painter
-        p.begin(self.picture)
         h = 1e-10
         for i in self.datasrc.df.index:
             prcr = prices[i]
@@ -1132,7 +1142,6 @@ class HorizontalTimeVolumeItem(CandlestickItem):
                 y = prcr[pocidx] + h / 2
                 p.setPen(pg.mkPen(poc_color))
                 p.drawLine(QtCore.QPointF(t, y), QtCore.QPointF(t+f*self.draw_poc, y))
-        p.end()
 
 
 
@@ -1229,6 +1238,14 @@ def create_plot_widget(master, rows=1, init_zoom_periods=1e10, yscale='linear'):
     return axs[0] if len(axs) == 1 else axs
 
 
+def close():
+    for win in windows:
+        try:
+            win.close()
+        except Exception as e:
+            print('Window closing error:', type(e), e)
+
+
 def price_colorfilter(item, datasrc, df):
     opencol = df.columns[1]
     closecol = df.columns[2]
@@ -1288,6 +1305,7 @@ def candlestick_ochl(datasrc, draw_body=True, draw_shadow=True, candle_width=0.6
     item = CandlestickItem(ax=ax, datasrc=datasrc, draw_body=draw_body, draw_shadow=draw_shadow, candle_width=candle_width, colorfunc=colorfunc)
     _update_significants(ax, datasrc, force=True)
     item.update_data = partial(_update_data, None, None, item)
+    item.update_gfx = partial(_update_gfx, item)
     ax.addItem(item)
     return item
 
@@ -1303,6 +1321,7 @@ def renko(x, y=None, bins=None, step=None, ax=None, colorfunc=price_colorfilter)
     item = candlestick_ochl(datasrc, draw_shadow=False, candle_width=1, ax=ax, colorfunc=colorfunc)
     item.colors['bull_body'] = item.colors['bull_frame']
     item.update_data = partial(_update_data, None, step_adjust_renko_datasrc, item)
+    item.update_gfx = partial(_update_gfx, item)
     global epoch_period
     epoch_period = (origdf.iloc[1,0] - origdf.iloc[0,0]) // int(1e9)
     return item
@@ -1326,6 +1345,7 @@ def volume_ocv(datasrc, candle_width=0.8, ax=None, colorfunc=volume_colorfilter)
         item.colors['weak_bull_frame'] = brighten(volume_bull_color, 1.2)
         item.colors['weak_bull_body']  = brighten(volume_bull_color, 1.2)
     item.update_data = partial(_update_data, None, _adjust_volume_datasrc, item)
+    item.update_gfx = partial(_update_gfx, item)
     ax.addItem(item)
     item.setZValue(-20)
     return item
@@ -1353,9 +1373,11 @@ def horiz_time_volume(datasrc, ax=None, **kwargs):
         datasrc.standalone = True # only force standalone if there is something on our charts already
     datasrc.scale_cols = [datasrc.col_data_offset, len(datasrc.df.columns)-2] # first and last price columns
     datasrc.pre_update = lambda df: df.loc[:, :df.columns[0]] # throw away previous data
+    datasrc.post_update = lambda df: df.dropna(how='all') # kill all-NaNs
     _set_datasrc(ax, datasrc)
     item = HorizontalTimeVolumeItem(ax=ax, datasrc=datasrc, **kwargs)
     item.update_data = partial(_update_data, _preadjust_horiz_datasrc, _adjust_horiz_datasrc, item)
+    item.update_gfx = partial(_update_gfx, item)
     item.setZValue(-10)
     ax.addItem(item)
     return item
@@ -1373,6 +1395,7 @@ def heatmap(datasrc, ax=None, **kwargs):
     _set_datasrc(ax, datasrc)
     item = HeatmapItem(ax=ax, datasrc=datasrc, **kwargs)
     item.update_data = partial(_update_data, None, None, item)
+    item.update_gfx = partial(_update_gfx, item)
     item.setZValue(-30)
     ax.addItem(item)
     if ax.vb.datasrc is not None and not ax.vb.datasrc.timebased(): # manual zoom update
@@ -1397,6 +1420,7 @@ def bar(x, y=None, width=0.8, ax=None, colorfunc=strength_colorfilter):
     _adjust_bar_datasrc(datasrc, order_cols=False) # don't rearrange columns, done for us in volume_ocv()
     item = volume_ocv(datasrc, candle_width=width, ax=ax, colorfunc=colorfunc)
     item.update_data = partial(_update_data, None, _adjust_bar_datasrc, item)
+    item.update_gfx = partial(_update_gfx, item)
     ax.vb.pre_process_data()
     if ax.vb.y_min >= 0:
         ax.vb.v_zoom_baseline = 0
@@ -1427,6 +1451,7 @@ def plot(x, y=None, color=None, width=1, ax=None, style=None, legend=None, zooms
     if style is None or any(ch in style for ch in '-_.'):
         connect_dots = 'finite' # same as matplotlib; use datasrc.standalone=True if you want to keep separate intervals on a plot
         item = ax.plot(x, y, pen=_makepen(color=used_color, style=style, width=width), name=legend, connect=connect_dots)
+        item.setZValue(5)
     else:
         symbol = {'v':'t', '^':'t1', '>':'t2', '<':'t3'}.get(style, style) # translate some similar styles
         yfilter = y.notnull()
@@ -1444,6 +1469,7 @@ def plot(x, y=None, color=None, width=1, ax=None, style=None, legend=None, zooms
     item.datasrc = datasrc
     _update_significants(ax, datasrc, force=False)
     item.update_data = partial(_update_data, None, None, item)
+    item.update_gfx = partial(_update_gfx, item)
     if ax.legend is not None:
         for _,label in ax.legend.items:
             if label.text == legend:
@@ -1461,6 +1487,7 @@ def labels(x, y=None, labels=None, color=None, ax=None, anchor=(0.5,1)):
     item = ScatterLabelItem(ax=ax, datasrc=datasrc, color=used_color, anchor=anchor)
     _update_significants(ax, datasrc, force=False)
     item.update_data = partial(_update_data, None, None, item)
+    item.update_gfx = partial(_update_gfx, item)
     ax.addItem(item)
     if ax.vb.v_zoom_scale > 0.9: # adjust to make hi/lo text fit
         ax.vb.v_zoom_scale = 0.9
@@ -1642,7 +1669,7 @@ def screenshot(file, fmt='png'):
         file.write(buffer.data())
         return True
     except Exception as e:
-        print(type(e), e)
+        print('Screenshot error:', type(e), e)
     return False
 
 
@@ -2014,7 +2041,7 @@ def _adjust_bar_datasrc(datasrc, order_cols=True):
     datasrc.scale_cols = [1, 2] # scale by both baseline and volume
 
 
-def _update_data(preadjustfunc, adjustfunc, item, ds):
+def _update_data(preadjustfunc, adjustfunc, item, ds, gfx=True):
     if preadjustfunc:
         ds = preadjustfunc(ds)
     ds = _create_datasrc(item.ax, ds)
@@ -2024,6 +2051,11 @@ def _update_data(preadjustfunc, adjustfunc, item, ds):
     if len(cs) >= len(ds.df.columns):
         ds.df.columns = cs[:len(ds.df.columns)]
     item.datasrc.update(ds)
+    if gfx:
+        item.update_gfx()
+
+
+def _update_gfx(item):
     _start_visual_update(item)
     for i in item.ax.items:
         if i == item or not hasattr(i, 'datasrc'):
@@ -2207,8 +2239,10 @@ def _inspect_pos(ax, inspector, poss):
             t = ax.vb.datasrc.x.iloc[-1 if t > 0 else 0]
     try:
         inspector(t, point.y())
+    except OSError as e:
+        pass
     except Exception as e:
-        print(type(e), e)
+        print('Inspection error:', type(e), e)
 
 
 def brighten(color, f):
@@ -2221,13 +2255,15 @@ def _get_color(ax, style, wanted_color):
     if type(wanted_color) in (str, QtGui.QColor):
         return wanted_color
     index = wanted_color if type(wanted_color) == int else None
-    if style is None or any(ch in style for ch in '-_.'):
+    is_line = lambda style: style is None or any(ch in style for ch in '-_.')
+    this_line = is_line(style)
+    if this_line:
         colors = soft_colors
     else:
         colors = hard_colors
     if index is None:
-        avoid = set(i.opts['handed_color'] for i in ax.items if isinstance(i,pg.PlotDataItem) and i.opts['handed_color'] is not None)
-        index = len([i for i in ax.items if isinstance(i,pg.PlotDataItem) and i.opts['handed_color'] is None])
+        avoid = set(i.opts['handed_color'] for i in ax.items if isinstance(i,pg.PlotDataItem) and i.opts['handed_color'] is not None and this_line==is_line(i.opts['symbol']))
+        index = len([i for i in ax.items if isinstance(i,pg.PlotDataItem) and i.opts['handed_color'] is None and this_line==is_line(i.opts['symbol'])])
         while index in avoid:
             index += 1
     return colors[index%len(colors)]
@@ -2350,12 +2386,12 @@ def _x2year(datasrc, x):
 
 
 def _round_to_significant(rng, rngmax, x, significant_decimals, significant_eps):
-    is_highres = (rng/significant_eps > 1e2 and rngmax<1e-2) or abs(rngmax) > 1e7
+    is_highres = (rng/significant_eps > 1e2 and rngmax<1e-2) or abs(rngmax) > 1e7 or rng < 1e-5
     sd = significant_decimals
     if is_highres and abs(x)>0:
         exp10 = floor(np.log10(abs(x)))
         x = x / (10**exp10)
-        sd = min(3, sd+int(np.log10(rngmax)))
+        sd = min(3, sd+int(abs(np.log10(rngmax))))
         fmt = '%%%i.%ife%%i' % (sd, sd)
         r = fmt % (x, exp10)
     else:
@@ -2384,8 +2420,9 @@ def _clamp_xy(ax, x, y):
             x = 0 if x < 0 else len(ds.df)-1
         x = round(x)
         eps = ax.significant_eps
-        eps2 = np.sign(y) * 0.5 * eps
-        y -= fmod(y+eps2, eps) - eps2
+        if eps > 1e-8:
+            eps2 = np.sign(y) * 0.5 * eps
+            y -= fmod(y+eps2, eps) - eps2
     y = ax.vb.yscale.invxform(y, verify=True)
     return x, y
 
