@@ -85,9 +85,7 @@ def update_candlestick_data(trade, interval_mins=1):
         return
     elif t > df['t'].iloc[-1]:
         # add new candle
-        o = df['c'].iloc[-1]
-        h = c if c>o else o
-        l = o if o<c else c
+        o = c = h = l = df['c'].iloc[-1]
         df1 = pd.DataFrame(dict(t=[t], o=[o], c=[c], h=[l], l=[l]))
         df = pd.concat([df, df1], ignore_index=True, sort=False)
     else:
@@ -112,6 +110,15 @@ def update_orderbook_data(orderbook10):
 
 if __name__ == '__main__':
     df = pd.DataFrame(price_history())
+
+    # fix bug in BitMEX websocket lib
+    def bugfix(self, *args, **kwargs):
+        from pyee import EventEmitter
+        EventEmitter.__init__(self)
+        orig_init(self, *args, **kwargs)
+    orig_init = Instrument.__init__
+    Instrument.__init__ = bugfix
+
     ws = Instrument(channels=[InstrumentChannels.trade, InstrumentChannels.orderBook10])
     @ws.on('action')
     def action(message):
